# coding: utf-8

from supervisely_lib.api.module_api import ApiField
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.api.entity_annotation.figure_api import FigureApi


class VideoFigureApi(FigureApi):
    def create(self, video_id, object_id, frame_index, geometry_json, geometry_type, track_id=None):
        return super().create(video_id,
                              object_id,
                              {ApiField.FRAME: frame_index},
                              geometry_json,
                              geometry_type,
                              track_id)

    def append_bulk(self, video_id, figures, key_id_map: KeyIdMap):
        keys = []
        figures_json = []
        for figure in figures:
            keys.append(figure.key())
            figures_json.append(figure.to_json(key_id_map, save_meta=True))

        self._append_bulk(video_id, figures_json, keys, key_id_map)




# coding: utf-8

from supervisely_lib._utils import take_with_default
from supervisely_lib.video_annotation.constants import FIGURES, INDEX
from supervisely_lib.video_annotation.video_figure import VideoFigure
from supervisely_lib.collection.key_indexed_collection import KeyObject


class Frame(KeyObject):
    '''
    This is a class for creating and using frames for videos
    '''
    def __init__(self, index, figures=None):
        '''

        :param index: int
        :param figures: list of figures(point, rectangle etc) on the frame
        '''
        self._index = index
        self._figures = take_with_default(figures, [])

    @property
    def index(self):
        return self._index

    def key(self):
        return self._index

    @property
    def figures(self):
        return self._figures.copy()

    def validate_figures_bounds(self, img_size=None):
        '''
        The function validate_figures_bounds checks if given image contains a figure. Raise error if figure is out of image bounds
        :param img_size: tuple of integers
        :return: None if image size is None
        '''
        if img_size is None:
            return
        for figure in self._figures:
            figure.validate_bounds(img_size, _auto_correct=False)

    def to_json(self, key_id_map=None):
        '''
        The function to_json convert frame to json format
        :param key_id_map: KeyIdMap class object
        :return: frame in json format
        '''
        data_json = {
            INDEX: self.index,
            FIGURES: [figure.to_json(key_id_map) for figure in self.figures]
        }
        return data_json

    @classmethod
    def from_json(cls, data, objects, frames_count=None, key_id_map=None):
        '''
        The function from_json convert frame from json format to Frame class object. Raise error if frame index < 0 and if frame index > number of frames in video
        :param data: input frame in json format
        :param objects: VideoObjectCollection
        :param frames_count: int
        :param key_id_map: KeyIdMap class object
        :return: Frame class object
        '''
        index = data[INDEX]
        if index < 0:
            raise ValueError("Frame Index have to be >= 0")

        if frames_count is not None:
            if index > frames_count:
                raise ValueError("Video contains {} frames. Frame index is {}".format(frames_count, index))

        figures = []
        for figure_json in data.get(FIGURES, []):
            figure = VideoFigure.from_json(figure_json, objects, index, key_id_map)
            figures.append(figure)
        return cls(index=index, figures=figures)

    def clone(self, index=None, figures=None):
        '''
        :param index: int
        :param figures: list
        :return: Frame class object
        '''
        return self.__class__(index=take_with_default(index, self.index),
                              figures=take_with_default(figures, self.figures))



# example from here
# https://stackoverflow.com/questions/38893448/pagination-on-pandas-dataframe-to-html/49134917
base_html = """
<!doctype html>
<html><head>
<meta http-equiv="Content-type" content="text/html; charset=utf-8">
<script type="text/javascript" src="https://ajax.googleapis.com/ajax/libs/jquery/2.2.2/jquery.min.js"></script>
<link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/1.10.16/css/jquery.dataTables.css">
<script type="text/javascript" src="https://cdn.datatables.net/1.10.16/js/jquery.dataTables.js"></script>
<link rel="stylesheet" href="https://fonts.googleapis.com/icon?family=Material+Icons">

<style id="compiled-css" type="text/css">
      * {box-sizing: border-box;}

.wrapper {
    border: 2px solid #f76707;
    border-radius: 5px;
    background-color: #fff4e6;
}

.wrapper > div {
    border: 2px solid #ffa94d;
    border-radius: 5px;
    background-color: #ffd8a8;
    padding: 1em;
    color: #d9480f;
}
.wrapper {
  display: grid;
  grid-template-columns: repeat(2, 1fr);
  grid-gap: 10px;
}

  </style>

</head><body>%s
</body></html>
"""

#<link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel="stylesheet">
#<script src="https://kit.fontawesome.com/a076d05399.js"></script>

table_html = """
<div>
    %s<script type="text/javascript">$(document).ready(function(){$('table').DataTable({"pageLength": 15 });});</script>
</div>
"""


div_wrapper = """
<div class="wrapper">{}</div>
"""

def df_html(df, index=True):
    return table_html % df.to_html(index=index, escape=False)


def compile_report(dfs, html_divs=[], index=True):
    result = ""
    for df in dfs:
        result += "{}\n".format(df_html(df, index))

    divs = ""
    for elem in html_divs:
        divs += "{}\n".format(elem)

    elmenets = div_wrapper.format(divs)
    result += elmenets
    return base_html % result




# coding: utf-8

import numpy as np

from supervisely_lib.sly_logger import logger
from supervisely_lib.metric.metric_base import MetricsBase
from supervisely_lib.metric.common import render_labels_for_class_name, safe_ratio, sum_counters, TOTAL_GROUND_TRUTH


INTERSECTION = 'intersection'
UNION = 'union'
IOU = 'iou'


def get_intersection(mask_1, mask_2):
    return (mask_1 & mask_2).sum()


def get_union(mask_1, mask_2):
    return (mask_1 | mask_2).sum()


def get_iou(mask_1, mask_2):
    return safe_ratio(get_intersection(mask_1, mask_2), get_union(mask_1, mask_2))


def _iou_log_line(iou, intersection, union):
    return 'IoU = {:.6f},  mean intersection = {:.6f}, mean union = {:.6f}'.format(iou, intersection, union)

def render_labels_as_binary_mask(labels, class_title, mask):
    for label in ann.labels:
        if label.obj_class.name == class_title:
                label.geometry.draw(mask, True)

class IoUMetric(MetricsBase):

    def __init__(self, class_mapping):
        self._class_mapping = class_mapping.copy()
        if len(self._class_mapping) < 1:
            raise RuntimeError('At least one classes pair should be defined.')
        self._counters = {cls_gt: {INTERSECTION: 0, UNION: 0, TOTAL_GROUND_TRUTH: 0}
                          for cls_gt in self._class_mapping.keys()}

    def add_pair(self, ann_gt, ann_pred):
        img_size = ann_gt.img_size
        for cls_gt, cls_pred in self._class_mapping.items():
            mask_gt, mask_pred = np.full(img_size, False), np.full(img_size, False)
            render_labels_as_binary_mask(ann_gt.labels, cls_gt, mask_gt)
            render_labels_as_binary_mask(ann_pred.labels, cls_pred, mask_pred)
            class_pair_counters = self._counters[cls_gt]
            class_pair_counters[INTERSECTION] += get_intersection(mask_gt, mask_pred)
            class_pair_counters[UNION] += get_union(mask_gt, mask_pred)
            class_pair_counters[TOTAL_GROUND_TRUTH] += 1

    def get_metrics(self):
        return {cls_gt: {INTERSECTION: safe_ratio(class_counters[INTERSECTION], class_counters[TOTAL_GROUND_TRUTH]),
                         UNION: safe_ratio(class_counters[UNION], class_counters[TOTAL_GROUND_TRUTH]),
                         IOU: safe_ratio(class_counters[INTERSECTION], class_counters[UNION])}
                for cls_gt, class_counters in self._counters.items()}

    def get_total_metrics(self):
        result = sum_counters(self._counters.values(), (INTERSECTION, UNION))
        result[IOU] = safe_ratio(result[INTERSECTION], result[UNION])
        return result

    def log_total_metrics(self):
        logger.info('**************** Result IoU metric values ****************')
        logger.info('NOTE! Values for "intersection" and "union" are in pixels.')
        for i, (cls_gt, values) in enumerate(self.get_metrics().items(), start=1):
            iou_line = _iou_log_line(values[IOU], values[INTERSECTION], values[UNION])
            logger.info('{}. Classes {} <-> {}:   {}'.format(i, cls_gt, self._class_mapping[cls_gt], iou_line))

        total_values = self.get_total_metrics()
        logger.info(
            'Total:   {}'.format(_iou_log_line(total_values[IOU], total_values[INTERSECTION], total_values[UNION])))



# coding: utf-8

from supervisely_lib.api.module_api import ApiField, ModuleApi
from supervisely_lib.io.fs import ensure_base_path
from supervisely_lib.imaging.image import read_bytes


class VideoFrameAPI(ModuleApi):
    def _download(self, video_id, frame_index):
        '''
        :param video_id: int
        :param frame_index: int
        :return: Response class object containing frame data with given index from given video id
        '''
        response = self._api.post('videos.download-frame', {ApiField.VIDEO_ID: video_id, ApiField.FRAME: frame_index})
        return response

    def download_np(self, video_id, frame_index):
        '''
        :param video_id: int
        :param frame_index: int
        :return: image in RGB format(numpy matrix) for frame with given index from given video id
        '''
        response = self._download(video_id, frame_index)
        img = read_bytes(response.content)
        return img

    def download_path(self, video_id, frame_index, path):
        '''
        :param video_id: int
        :param frame_index: int
        :param path: str
        :return: write image on the given path for frame with given index from given video id
        '''
        response = self._download(video_id, frame_index)
        ensure_base_path(path)
        with open(path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)



# coding: utf-8

import uuid

from supervisely_lib.video_annotation.video_object import VideoObject
from supervisely_lib.annotation.label import LabelJsonFields
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.video_annotation.constants import KEY, ID
from supervisely_lib.video_annotation.video_tag_collection import VideoTagCollection
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, CLASS_ID


class PointcloudObject(VideoObject):
    '''
    This is a class for creating and using PointcloudObject objects for videos
    '''
    @classmethod
    def from_json(cls, data, project_meta: ProjectMeta, key_id_map: KeyIdMap = None):
        '''
        The function from_json convert PointcloudObject from json format to PointcloudObject class object. Raise error if object class name is not found in the given project meta
        :param data: input PointcloudObject in json format
        :param project_meta: ProjectMeta class object
        :param key_id_map: KeyIdMap class object
        :return: PointcloudObject class object
        '''
        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(f'Failed to deserialize a object from JSON: class name {obj_class_name!r} '
                               f'was not found in the given project meta.')

        object_id = data.get(ID, None)

        existing_key = None
        if object_id is not None and key_id_map is not None:
            existing_key = key_id_map.get_object_key(object_id)
        json_key = uuid.UUID(data[KEY]) if KEY in data else None
        if (existing_key is not None) and (json_key is not None) and (existing_key != json_key):
            raise RuntimeError("Object id = {!r}: existing_key {!r} != json_key {!r}"
                               .format(object_id, existing_key, json_key))

        if existing_key is not None:
            key = existing_key
        elif json_key is not None:
            key = json_key
        else:
            key = uuid.uuid4()

        if key_id_map is not None and existing_key is None:
            key_id_map.add_object(key, object_id)

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(obj_class=obj_class,
                   key=key,
                   tags=VideoTagCollection.from_json(data[LabelJsonFields.TAGS], project_meta.tag_metas),
                   class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)


# coding: utf-8

from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.validation import is_2d_int_coords_valid
from supervisely_lib.geometry.sliding_windows import SlidingWindows
from supervisely_lib.collection.str_enum import StrEnum


class SlidingWindowBorderStrategy(StrEnum):
    ADD_PADDING = 'add_padding'
    SHIFT_WINDOW = 'shift_window'
    CHANGE_SIZE = 'change_size'


class SlidingWindowsFuzzy(SlidingWindows):
    def __init__(self, window_shape, min_overlap, strategy=str(SlidingWindowBorderStrategy.SHIFT_WINDOW)):
        super().__init__(window_shape, min_overlap)
        if not SlidingWindowBorderStrategy.has_value(strategy):
            raise ValueError("Unknown strategy {!r}. Allowed strategies: {}"
                             .format(strategy, SlidingWindowBorderStrategy.values()))
        self.strategy = strategy

    def get(self, source_shape):
        if self.strategy == str(SlidingWindowBorderStrategy.SHIFT_WINDOW):
            yield from self.get_shift_window(source_shape)
        elif self.strategy == str(SlidingWindowBorderStrategy.ADD_PADDING):
            yield from self.get_add_padding(source_shape)
        elif self.strategy == str(SlidingWindowBorderStrategy.CHANGE_SIZE):
            yield from self.get_change_size(source_shape)
        else:
            raise NotImplementedError("Not implemented SW Strategy: {!r}".format(self.strategy))

    def get_shift_window(self, source_shape):
        yield from super().get(source_shape)

    def get_add_padding(self, source_shape):
        source_rect = Rectangle.from_size(source_shape)
        window_rect = Rectangle.from_size(self.window_shape)
        if not source_rect.contains(window_rect):
            raise RuntimeError('Sliding window: window is larger than source (image).')

        hw_limit = tuple(source_shape[i] - self.window_shape[i] for i in (0, 1))
        for wind_top in range(0, hw_limit[0] + self.stride[0], self.stride[0]):
            for wind_left in range(0, hw_limit[1] + self.stride[1], self.stride[1]):
                roi = window_rect.translate(drow=wind_top, dcol=wind_left)
                yield roi

    def get_change_size(self, source_shape):
        source_rect = Rectangle.from_size(source_shape)
        window_rect = Rectangle.from_size(self.window_shape)
        if not source_rect.contains(window_rect):
            raise RuntimeError('Sliding window: window is larger than source (image).')

        hw_limit = tuple(source_shape[i] - self.window_shape[i] for i in (0, 1))
        for wind_top in range(0, hw_limit[0] + self.stride[0], self.stride[0]):
            for wind_left in range(0, hw_limit[1] + self.stride[1], self.stride[1]):
                wind_bottom = min(wind_top + self.stride[0], source_shape[0])
                wind_right = min(wind_left + self.stride[1], source_shape[1])
                roi = Rectangle(wind_top, wind_left, wind_bottom-1, wind_right-1)
                if not source_rect.contains(roi):
                    raise RuntimeError('Sliding window: result crop bounds are invalid.')
                yield roi



from supervisely_lib.nn.dataset import SlyDataset

from torchvision.transforms.functional import to_tensor


class PytorchSegmentationSlyDataset(SlyDataset):
    """Thin wrapper around the base Supervisely dataset IO logic to handle PyTorch specific conversions.

    The base class handles locating and reading images and annotation data from disk, conversions between named classes
    and integer class ids to feed to the models, and rendering annotations as images.
    """

    def _get_sample_impl(self, img_fpath, ann_fpath):
        # Read the image, read the annotation, render the annotation as a bitmap of per-pixel class ids, resize to
        # requested model input size.
        img, gt = super()._get_sample_impl(img_fpath=img_fpath, ann_fpath=ann_fpath)

        # PyTorch specific logic:
        # - convert from (H x W x Channels) to (Channels x H x W);
        # - convert from uint8 to float32 data type;
        # - move from 0...255 to 0...1 intensity range.
        img_tensor = to_tensor(img)

        return img_tensor, gt


from .compare_gallery import CompareGallery
from .chart import Chart
from .empty_grid_gallery import EmptyGridGallery
from .predictions_dynamics_gallery import PredictionsDynamicsGallery
from .progress_bar import ProgressBar
from .single_image_gallery import SingleImageGallery



# coding: utf-8
import random
import numpy as np
import imgaug.augmenters as iaa

from supervisely_lib.imaging import image as sly_image
from supervisely_lib.annotation.annotation import Annotation
from supervisely_lib.geometry.image_rotator import ImageRotator
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib._utils import take_with_default
from supervisely_lib.sly_logger import logger


def _validate_image_annotation_shape(img: np.ndarray, ann: Annotation) -> None:
    if img.shape[:2] != ann.img_size:
        raise RuntimeError('Image shape {} doesn\'t match img_size {} in annotation.'.format(
            img.shape[:2], ann.img_size))


# Flips
def fliplr(img: np.ndarray, ann: Annotation) -> (np.ndarray, Annotation):
    """
    Flips an image array and annotation around vertical axis.

    Args:
        img: Input image array.
        ann: Input annotation.
    Returns:
        A tuple containing flipped image and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    res_img = sly_image.fliplr(img)
    res_ann = ann.fliplr()
    return res_img, res_ann


def flipud(img: np.ndarray, ann: Annotation) -> (np.ndarray, Annotation):
    """
    Flips an image array and annotation around horizontal axis.

    Args:
        img: Input image array.
        ann: Input annotation.
    Returns:
        A tuple containing flipped image and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    res_img = sly_image.flipud(img)
    res_ann = ann.flipud()
    return res_img, res_ann


# Crops
def crop(img: np.ndarray, ann: Annotation, top_pad: int = 0, left_pad: int = 0, bottom_pad: int = 0,
         right_pad: int = 0) -> (np.ndarray, Annotation):
    """
    Crops the given image array and annotation from all sides with the given values.

    Args:
        img: Input image array.
        ann: Input annotation.
        top_pad: The size in pixels of the piece of picture that will be cut from the top side.
        left_pad: The size in pixels of the piece of picture that will be cut from the left side.
        bottom_pad: The size in pixels of the piece of picture that will be cut from the bottom side.
        right_pad: The size in pixels of the piece of picture that will be cut from the right side.
    Returns:
        A tuple containing cropped image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    height, width = img.shape[:2]
    crop_rect = Rectangle(top_pad, left_pad, height - bottom_pad - 1, width - right_pad - 1)

    res_img = sly_image.crop(img, crop_rect)
    res_ann = ann.relative_crop(crop_rect)
    return res_img, res_ann


def crop_fraction(img: np.ndarray, ann: Annotation, top: float = 0, left: float = 0, bottom: float = 0,
                  right: float = 0) -> (np.ndarray, Annotation):
    """
        Crops the given image array and annotation from all sides with the given fraction values.

        Args:
            img: Input image array.
            ann: Input annotation.
            top: The relative size of the piece of picture that will be cut from the top side.
            left: The relative size  of the piece of picture that will be cut from the left side.
            bottom: The relative size  of the piece of picture that will be cut from the bottom side.
            right: The relative size  of the piece of picture that will be cut from the right side.
        Returns:
            A tuple containing cropped image array and annotation.
        """
    _validate_image_annotation_shape(img, ann)
    if not all(0 <= pad < 1 for pad in (top, left, right, bottom)):
        raise ValueError('All padding values must be between 0 and 1.')
    height, width = img.shape[:2]
    top_pixels = round(height * top)
    left_pixels = round(width * left)
    bottom_pixels = round(height * bottom)
    right_pixels = round(width * right)
    return crop(img, ann, top_pad=top_pixels, left_pad=left_pixels, bottom_pad=bottom_pixels,
                right_pad=right_pixels)


def random_crop(img: np.ndarray, ann: Annotation, height: int, width: int) -> (np.ndarray, Annotation):
    """
    Crops given image array and annotation at a random location.

    Args:
        img: Input image array.
        ann: Input annotation.
        height: Desired height of output crop.
        width: Desired width of output crop.
    Returns:
        A tuple containing cropped image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    img_height, img_width = img.shape[:2]

    def calc_crop_pad(old_side, crop_side):
        new_side = min(int(old_side), int(crop_side))
        min_bound = random.randint(0, old_side - new_side)  # including [a; b]
        max_bound = old_side - min_bound - new_side
        return min_bound, max_bound

    left_pad, right_pad = calc_crop_pad(img_width, width)
    top_pad, bottom_pad = calc_crop_pad(img_height, height)
    return crop(img, ann, top_pad=top_pad, left_pad=left_pad, bottom_pad=bottom_pad, right_pad=right_pad)


def random_crop_fraction(
        img: np.ndarray, ann: Annotation, height_fraction_range: tuple, width_fraction_range: tuple) -> \
        (np.ndarray, Annotation):
    """
    Crops given image array and annotation at a random location with random size lying in the set intervals.

    Args:
        img: Input image array.
        ann: Input annotation
        height_fraction_range: Range of relative values [0, 1] to select output height from.
        width_fraction_range: Range of relative values [0, 1] to select output width from.
    Returns:
         A tuple containing cropped image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    img_height, img_width = img.shape[:2]

    height_p = random.uniform(height_fraction_range[0], height_fraction_range[1])
    width_p = random.uniform(width_fraction_range[0], width_fraction_range[1])
    crop_height = round(img_height * height_p)
    crop_width = round(img_width * width_p)
    return random_crop(img, ann, height=crop_height, width=crop_width)


def batch_random_crops_fraction(
        img_ann_pairs: list, crops_per_image: int, height_fraction_range: tuple, width_fraction_range: tuple) -> list:
    return [random_crop_fraction(img, ann, height_fraction_range, width_fraction_range)
            for img, ann in img_ann_pairs for _ in range(crops_per_image)]


def flip_add_random_crops(
        img: np.ndarray, ann: Annotation, crops_per_image: int, height_fraction_range: tuple, width_fraction_range: tuple) -> list:
    full_size_items = [(img, ann), fliplr(img, ann)]
    crops = batch_random_crops_fraction(full_size_items, crops_per_image, height_fraction_range, width_fraction_range)
    return full_size_items + crops


# TODO factor out / simplify.
def _rect_from_bounds(padding_config: dict, img_h: int, img_w: int) -> Rectangle:
    def get_padding_pixels(raw_side, dim_name):
        side_padding_config = padding_config.get(dim_name)
        if side_padding_config is None:
            padding_pixels = 0
        elif side_padding_config.endswith('px'):
            padding_pixels = int(side_padding_config[:-len('px')])
        elif side_padding_config.endswith('%'):
            padding_fraction = float(side_padding_config[:-len('%')])
            padding_pixels = int(raw_side * padding_fraction / 100.0)
        else:
            raise ValueError(
                'Unknown padding size format: {}. Expected absolute values as "5px" or relative as "5%"'.format(
                    side_padding_config))
        return padding_pixels

    def get_padded_side(raw_side, l_name, r_name):
        l_bound = -get_padding_pixels(raw_side, l_name)
        r_bound = raw_side + get_padding_pixels(raw_side, r_name)
        return l_bound, r_bound

    left, right = get_padded_side(img_w, 'left', 'right')
    top, bottom = get_padded_side(img_h, 'top', 'bottom')
    return Rectangle(top=top, left=left, bottom=bottom, right=right)


def instance_crop(img: np.ndarray, ann: Annotation, class_title: str, save_other_classes_in_crop: bool = True,
                  padding_config: dict = None) -> list:
    """
    Crops objects of specified classes from image with configurable padding.

    Args:
        img: Input image array.
        ann: Input annotation.
        class_title: Name of class to crop.
        save_other_classes_in_crop: save non-target classes in each cropped annotation.
        padding_config: Dict with padding
    Returns:
        List of cropped [image, annotation] pairs.
    """
    padding_config = take_with_default(padding_config, {})
    _validate_image_annotation_shape(img, ann)
    results = []
    img_rect = Rectangle.from_size(img.shape[:2])

    if save_other_classes_in_crop:
        non_target_labels = [label for label in ann.labels if label.obj_class.name != class_title]
    else:
        non_target_labels = []

    ann_with_non_target_labels = ann.clone(labels=non_target_labels)

    for label in ann.labels:
        if label.obj_class.name == class_title:
            src_fig_rect = label.geometry.to_bbox()
            new_img_rect = _rect_from_bounds(padding_config, img_w=src_fig_rect.width, img_h=src_fig_rect.height)
            rect_to_crop = new_img_rect.translate(src_fig_rect.top, src_fig_rect.left)
            crops = rect_to_crop.crop(img_rect)
            if len(crops) == 0:
                continue
            rect_to_crop = crops[0]
            image_crop = sly_image.crop(img, rect_to_crop)

            cropped_ann = ann_with_non_target_labels.relative_crop(rect_to_crop)

            label_crops = label.relative_crop(rect_to_crop)
            for label_crop in label_crops:
                results.append((image_crop, cropped_ann.add_label(label_crop)))
    return results


# Resize
def resize(img: np.ndarray, ann: Annotation, size: tuple) -> (np.ndarray, Annotation):
    """
    Resize the input image array and annotation to the given size.

    Args:
        img: Input image array.
        ann: Input annotation.
        size: Desired size (height, width) in pixels or -1. If one of values is -1 and "keep": true then for
                specific width height will be automatically computed to keep aspect ratio.
    Returns:
        A tuple containing resized image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    height = take_with_default(size[0], -1)  # For backward capability
    width = take_with_default(size[1], -1)
    size = (height, width)

    new_size = sly_image.restore_proportional_size(in_size=ann.img_size, out_size=size)
    res_img = sly_image.resize(img, new_size)
    res_ann = ann.resize(new_size)
    return res_img, res_ann


# Resize
def scale(img: np.ndarray, ann: Annotation, frow: float = None, fcol: float = None, f: float = None) \
        -> (np.ndarray, Annotation):
    """
    Resize the input image array and annotation to the given size.

    Args:
        img: Input image array.
        ann: Input annotation.
        frow: Desired height scale height value
        frow: Desired width scale width value
        f: Desired height and width scale values in one
    Returns:
        A tuple containing resized image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    new_size = sly_image.restore_proportional_size(in_size=ann.img_size, frow=frow, fcol=fcol, f=f)
    res_img = sly_image.resize(img, new_size)
    res_ann = ann.resize(new_size)
    return res_img, res_ann


# Rotate
class RotationModes:
    KEEP = 'keep'
    CROP = 'crop'


def rotate(img: np.ndarray, ann: Annotation, degrees: float, mode: str=RotationModes.KEEP) ->\
        (np.ndarray, Annotation):  # @TODO: add "preserve_size" mode
    """
    Rotates the image by random angle.

    Args:
        img: Input image array.
        ann: Input annotation.
        degrees: Rotation angle, counter-clockwise.
        mode: parameter: "keep" - keep original image data, then new regions will be filled with black color;
            "crop" - crop rotated result to exclude black regions;
    Returns:
        A tuple containing rotated image array and annotation.
    """
    _validate_image_annotation_shape(img, ann)
    rotator = ImageRotator(img.shape[:2], degrees)

    if mode == RotationModes.KEEP:
        rect_to_crop = None

    elif mode == RotationModes.CROP:
        rect_to_crop = rotator.inner_crop

    else:
        raise NotImplementedError('Wrong black_regions mode.')

    res_img = rotator.rotate_img(img, use_inter_nearest=False)
    res_ann = ann.rotate(rotator)
    if rect_to_crop is not None:
        res_img = sly_image.crop(res_img, rect_to_crop)
        res_ann = res_ann.relative_crop(rect_to_crop)
    return res_img, res_ann


def load_imgaug(json_data):
    def _get_function(category_name, aug_name):
        try:
            submodule = getattr(iaa, category_name)
            aug_f = getattr(submodule, aug_name)
            return aug_f
        except Exception as e:
            logger.error(repr(e))
            raise e

    pipeline_json = json_data["pipeline"]
    random_order = json_data.get("random_order", False)

    pipeline = []
    for aug_info in pipeline_json:
        category_name = aug_info["category"]
        aug_name = aug_info["name"]
        params = aug_info["params"]

        aug_func = _get_function(category_name, aug_name)
        aug = aug_func(**params)

        sometimes = aug_info.get("sometimes", None)
        if sometimes is not None:
            aug = iaa.meta.Sometimes(sometimes, aug)
        pipeline.append(aug)
    augs = iaa.Sequential(pipeline, random_order=random_order)
    return augs




# coding: utf-8

import json
from supervisely_lib.api.module_api import ApiField
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.video_annotation.video_annotation import VideoAnnotation

from supervisely_lib.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely_lib.io.json import load_json_file


class VideoAnnotationAPI(EntityAnnotationAPI):
    _method_download_bulk = 'videos.annotations.bulk.info'
    _entity_ids_str = ApiField.VIDEO_IDS

    def download(self, video_id):
        '''
        :param video_id: int
        :return: video annotation to given id in json format
        '''
        video_info = self._api.video.get_info_by_id(video_id)
        return self._download(video_info.dataset_id, video_id)

    def append(self, video_id, ann: VideoAnnotation, key_id_map: KeyIdMap = None):
        '''
        ???
        :param video_id: int
        :param ann: VideoAnnotation class object
        :param key_id_map: KeyIdMap class object
        '''
        info = self._api.video.get_info_by_id(video_id)
        self._append(self._api.video.tag, self._api.video.object, self._api.video.figure,
                     info.project_id, info.dataset_id, video_id,
                     ann.tags, ann.objects, ann.figures, key_id_map)

    def upload_paths(self, video_ids, ann_paths, project_meta, progress_cb=None):
        # video_ids from the same dataset

        for video_id, ann_path in zip(video_ids, ann_paths):
            ann_json = load_json_file(ann_path)
            ann = VideoAnnotation.from_json(ann_json, project_meta)

            # ignore existing key_id_map because the new objects will be created
            self.append(video_id, ann)


# coding: utf-8

from enum import Enum
from supervisely_lib.api.module_api import ApiField, ModuleApi, ModuleWithStatus


class AgentNotFound(Exception):
    pass


class AgentNotRunning(Exception):
    pass


class AgentApi(ModuleApi, ModuleWithStatus):
    class Status(Enum):
        WAITING = 'waiting'
        RUNNING = 'running'

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.TOKEN,
                ApiField.STATUS,
                ApiField.USER_ID,
                ApiField.TEAM_ID,
                ApiField.CAPABILITIES,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'AgentInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(self, team_id, filters=None):
        '''
        :param team_id: int
        :param filters: list
        :return: list of all agents with given team id
        '''
        return self.get_list_all_pages('agents.list',  {'teamId': team_id, "filter": filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: workspace metadata by workspace numeric ID
        '''
        return self._get_info_by_id(id, 'agent.info')

    def get_status(self, id):
        '''
        :param id: int
        :return: Status class object containing status of workspace with given numeric ID (maybe waiting or running)
        '''
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        pass



# coding: utf-8

from copy import deepcopy
from typing import List
from supervisely_lib.imaging.color import random_rgb, rgb2hex, hex2rgb, _validate_color
from supervisely_lib.io.json import JsonSerializable
from supervisely_lib.collection.key_indexed_collection import KeyObject
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.any_geometry import AnyGeometry
from supervisely_lib._utils import take_with_default
from supervisely_lib.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR


class ObjClassJsonFields:
    ID = 'id'
    NAME = 'title'
    GEOMETRY_TYPE = 'shape'
    COLOR = 'color'
    GEOMETRY_CONFIG = 'geometry_config'
    HOTKEY = 'hotkey'


class ObjClass(KeyObject, JsonSerializable):
    def __init__(self, name: str, geometry_type: type, color: List[int]=None, geometry_config: dict=None, sly_id=None,
                 hotkey: str = None):
        """
        Class of objects (person, car, etc) with necessary properties: name, type of geometry (Polygon, Rectangle, ...)
        and RGB color. Only one class can be associated with Label.


        Args:
            name: string name of the class (person, car, apple, etc)
            geometry_type: type of the geometry. Geometry defines the shape for all Labels of this ObjClass:
                Polygon, Rectangle, Bitmap, Polyline, Point
            color: [R, G, B]
            geometry_config: additional settings of the geometry that is associated with ObjClass
        Returns:
            ObjClass instance
        """
        self._name = name
        self._geometry_type = geometry_type
        self._color = random_rgb() if color is None else deepcopy(color)
        self._geometry_config = deepcopy(take_with_default(geometry_config, {}))
        self._sly_id = sly_id
        self._hotkey = take_with_default(hotkey, "")
        _validate_color(self._color)

    @property
    def name(self):
        """
        Returns:
            string name of the ObjectClass
        """
        return self._name

    def key(self):
        """
        Used as a key in ObjClassCollection (like key in dict)
        Returns:
            string name of the ObjectClass
        """
        return self.name

    @property
    def geometry_type(self):
        """
        Returns:
            type of the geometry that is associated with ObjClass (Polygon, Rectangle, Bitmap, etc)
        """
        return self._geometry_type

    @property
    def geometry_config(self):
        """
        Returns:
            additional settings of the geometry that is associated with ObjClass
        """
        return deepcopy(self._geometry_config)

    @property
    def color(self):
        """
        Returns:
            [R, G, B]
        """
        return deepcopy(self._color)

    @property
    def sly_id(self):
        return self._sly_id

    @property
    def hotkey(self):
        return self._hotkey

    def to_json(self) -> dict:
        """
        Converts object to json serializable dictionary. See Supervisely Json format explanation here:
        https://docs.supervise.ly/ann_format/

        Returns:
            json serializable dictionary
        """
        res = {
            ObjClassJsonFields.NAME: self.name,
            ObjClassJsonFields.GEOMETRY_TYPE: self.geometry_type.geometry_name(),
            ObjClassJsonFields.COLOR: rgb2hex(self.color),
            ObjClassJsonFields.GEOMETRY_CONFIG: self.geometry_type.config_to_json(self._geometry_config)
        }
        if self.sly_id is not None:
            res[ObjClassJsonFields.ID] = self.sly_id
        if self._hotkey is not None:
            res[ObjClassJsonFields.HOTKEY] = self.hotkey
        return res

    @classmethod
    def from_json(cls, data: dict) -> 'ObjClass':
        """
        Creates object from json serializable dictionary. See Supervisely Json format explanation here:
        https://docs.supervise.ly/ann_format/

        Returns:
            ObjClass
        """
        name = data[ObjClassJsonFields.NAME]
        geometry_type = GET_GEOMETRY_FROM_STR(data[ObjClassJsonFields.GEOMETRY_TYPE])
        color = hex2rgb(data[ObjClassJsonFields.COLOR])
        geometry_config = geometry_type.config_from_json(data.get(ObjClassJsonFields.GEOMETRY_CONFIG))
        sly_id = data.get(ObjClassJsonFields.ID, None)
        hotkey = data.get(ObjClassJsonFields.HOTKEY, "")
        return cls(name=name, geometry_type=geometry_type, color=color, geometry_config=geometry_config, sly_id=sly_id,
                   hotkey=hotkey)

    def __eq__(self, other: 'ObjClass'):
        return isinstance(other, ObjClass) and \
               self.name == other.name and \
               (self.geometry_type == other.geometry_type or AnyGeometry in [self.geometry_type, other.geometry_type]) and \
               self.geometry_config == other.geometry_config

    def __ne__(self, other: 'ObjClass'):
        return not self == other

    def __str__(self):  # Is need show geometry settings here?
        return '{:<7s}{:<10}{:<7s}{:<13}{:<7s}{:<15}{:<16s}{:<16}{:<7s}{:<7}'.format(
            'Name:', self.name,
            'Shape:', self.geometry_type.__name__,
            'Color:', str(self.color),
            'Geom. settings:', str(self.geometry_config),
            'Hotkey', self.hotkey
        )

    @classmethod
    def get_header_ptable(cls):
        return ['Name', 'Shape', 'Color', 'Hotkey']  # Is need show geometry settings here?

    def get_row_ptable(self):
        return [self.name, self.geometry_type.__name__, self.color, self.hotkey]

    def clone(self, name: str = None, geometry_type: Geometry = None, color: List[int] = None,
              geometry_config: dict = None, sly_id=None, hotkey=None) -> 'ObjClass':
        """
        Creates object duplicate. Defined arguments replace corresponding original values.

        Args:
            see __init__ method
        Returns:
            new instance of class
        """
        return ObjClass(name=take_with_default(name, self.name),
                        geometry_type=take_with_default(geometry_type, self.geometry_type),
                        color=take_with_default(color, self.color),
                        geometry_config=take_with_default(geometry_config, self.geometry_config),
                        sly_id=take_with_default(sly_id, self.sly_id),
                        hotkey=take_with_default(hotkey, self.hotkey))

    def __hash__(self):
        return hash((self.name, self.geometry_type.geometry_name()))